*************************************************************************************
*														*
*	Matrix character simulation for the EASy68k simulator	2007/04/19			*
*														*
*	This code uses double buffering to render a patterm similar to that seen	*
*	on tank's screens in 'The Matrix' series of films.					*
*														*
*	The F2, F3 and F4	keys can be used to select a screen size of 640 x 480,	*
*	800 x 600 and 1024 x 768 respectively.							*
*														*
*	This version for Sim68K 3.6 or later							*
*														*
*	More 68000 and other projects can be found on my website at ..			*
*														*
*	 http://members.lycos.co.uk/leeedavison/index.html					*
*														*
*	mail : leeedavison@lycos.co.uk								*
*														*
*************************************************************************************

tail		EQU	$0E				* length of tail that is not erased
erase		EQU	$09				* erase character step, the smaller this
							* number the faster the column is cleared
spawn		EQU	$1F				* chance of new spawn is 1 in (n-1) where n
							* is an integer power of two
colours	EQU	$0F				* number of colours for tails, if this is
							* changed the palette must be changed to match

	ORG	$1000

start
	MOVEQ		#17,d1			* enable double buffering
	MOVEQ		#92,d0			* set draw mode
	TRAP		#15

	MOVEQ		#12,d1			* background AND pen colour
	MOVEQ		#92,d0			* set draw mode
	TRAP		#15

	BSR		init				* go initialise everything

	MOVE.w	#$FF00,d1			* clear screen
	MOVEQ		#11,d0			* position cursor
	TRAP		#15

main_loop
	MOVEQ		#94,d0			* copy screen buffer to main
	TRAP		#15

	MOVEQ		#8,d0				* get time in 1/100 ths seconds
	TRAP		#15

	MOVE.l	d1,-(sp)			* push the time on the stack

	BSR		screen_size			* test and handle widow size change keypresses

	MOVEQ		#0,d7				* clear longword
	MOVE.w	screen_width(a3),d7	* get the screen width
	DIVU		#8,d7				* calculate the number of columns
	SUBQ.w	#1,d7				* set the index to the end column
column_loop
	MOVE.w	d7,d0				* copy the column index
	ADD.w		d0,d0				* * 2 bytes per row(column%)
	MOVEQ		#0,d6				* y% = 0, clear word
	MOVE.b	(a6,d0.w),d6		* y% = INT(row(column%))

	CMPI.b	#$FE,d6			* compare y% with - 2
	BNE		draw_string			* if <> - 2 go draw the string

	BSR		spawn_new			* possibly spawn a new string
	BRA.s		loop_end			* go do the column loop tail

draw_string
	MOVE.w	(a5,d0.w),d5		* get the deltav(column%)
	ADD.w		d5,(a6,d0.w)		* row(column%) = row(column%) + deltav(column%)

	MOVEQ		#0,d5				* next y% = 0, clear word
	MOVE.b	(a6,d0.w),d5		* next y% = INT(row(column%))

	BSR		print_tail			* draw the tail character
	BSR		print_head			* draw the head character
loop_end
	BSR		erase_chr			* erase a character from the current column
	DBF		d7,column_loop		* decrement column and loop if more to do

* now wait for a while. if you have a slow machine this shouldn't slow it down much but
* if you have a very fast machine them this will stop the effect running too fast

	MOVE.l	(sp)+,d7			* restore the time in 1/100ths from stack
wait_100ms
	MOVEQ		#8,d0				* get the current time in 1/100ths seconds
	TRAP		#15

	SUB.l		d7,d1				* subtract the previous time from the current
	CMP.b		#$09,d1			* compare difference with 9/100ths of a second
	BMI.s		wait_100ms			* loop if time is not up yet

* if you want to see how fast it can go comment out the BMI.s above.

	BRA		main_loop			* go round forever


*************************************************************************************
*
* check the [F2], [F3] and [F4] keys. set the screen size to 640 x 480, 800 x 600 or
* 1024 x 768 if the corresponding key has been pressed

screen_size
	MOVE.l	#$71007273,d1		* [F2], [], [F3] and [F4] keys
	MOVEQ		#19,d0			* check for keypress
	TRAP		#15

	MOVE.l	d1,d2				* copy result
	BEQ.s		notscreen			* skip screen size if no F key

	MOVE.l	#$028001E0,d1		* set 640 x 480
	TST.l		d2				* test result
	BMI		setscreen			* if F2 go set window size

	MOVE.l	#$03200258,d1		* set 800 x 600
	TST.w		d2				* test result
	BMI		setscreen			* if F3 go set window size

							* else was F4 so ..
	MOVE.l	#$04000300,d1		* set 1024 x 768
setscreen
	CMP.l		screen_width(a3),d1	* compare with current screen size
	BEQ.s		notscreen			* if already set skip setting it now

	BRA		restart_screen		* go reset all the columns and change the
							* window size

notscreen
	RTS


*************************************************************************************
*
* possibly spawn a new string as the last one is done

spawn_new
	BSR		random_n			* generate a random number
	MOVEQ		#0,d0				* clear high word
	MOVE.w	2(a3),d0			* get a random word
	DIVU.w	#spawn,d0			* divide by spawn rate
	SWAP		d0				* get remainder
	TST.w		d0				* test the remainder
	BNE		exit_spawn			* if not zero skip spawnimg a new string

	MOVE.w	d7,d0				* copy column
	ADD.w		d0,d0				* * 2
	MOVE.w	#$FF00,(a6,d0.w)		* row(column%) = - 1.0

	MOVEQ		#$40,d1			* set minimum delta v
	OR.b		1(a3),d1			* add random byte
	MOVE.w	d1,(a5,d0.w)		* deltav(column%) = (RND * 1) + 1/4

	BSR		random_n			* generate random number
	MOVEQ		#0,d1				* clear longword
	MOVE.w	(a3),d1			* get random word
	DIVU.w	#colours,d1			* divide by the number of colours
	SWAP		d1				* use remainder, 0 to number - 1
	MOVE.b	d1,(a4,d7.w)		* colour.b(column%) = (RND * colours)

exit_spawn
	RTS


*************************************************************************************
*
* draw the tail character

print_tail
	TST.b		d6				* test y%
	BMI		exit_tail			* if y% < 0 just exit

	BSR		get_rows			* get the number of screen rows in d1.w
	ADD.w		#tail,d1			* add the tail length
	CMP.b		d1,d6				* compare y% with screen rows + tail
	BCC		no_tail			* if y% >= screen rows + tail flag done

	BSR		get_rows			* get the number of screen rows in d1.w
	CMP.b		d1,d6				* compare y% with the number of screen rows
	BCC		exit_tail			* if y% >= the number of screen rows just exit

	CMP.b		d6,d5				* compare % with next y%
	BEQ		exit_tail			* if y% = next y% just exit

	MOVE.w	d6,d2				* copy y%
	BSR		set_cursor			* set the text cursor to (column%, d2)

	MOVEQ		#0,d4				* clear longword
	MOVE.b	(a4,d7.w),d4		* get column colour
	BRA		colour_character		* go print the character and colour it

no_tail
	MOVE.w	#$FE00,(a6,d0.w)		* else flag done, make row(column%) = - 2.0
exit_tail
	RTS


*************************************************************************************
*
* draw the head character

print_head
	TST.b		d5				* test next y%
	BMI		no_head			* if next y% < 0 just exit
	
	BSR		get_rows			* get the number of screen rows in d1.w
	CMP.b		d1,d5				* compare next y% with the number of screen
							* rows
	BCC		no_head			* if next y% >= the number of screen rows
							* just exit

	MOVE.w	d5,d2				* copy next y%
	BSR		set_cursor			* set the text cursor to (column%, d2)

	MOVEQ		#colours,d4			* set colour to head colour
	BRA		colour_character		* go print the character and colour it

no_head
	RTS


*************************************************************************************
*
* erase a random character from the current column

erase_chr
	MOVEQ		#0,d1				* set $00000000, black fill
	MOVEQ		#81,d0			* set the fill colour
	TRAP		#15				*

	BSR		random_n			* generate random number
	MOVEQ		#0,d4				* clear longword
	MOVE.w	(a3),d4			* get a random longword
	DIVU.w	#erase,d4			* divide by erase step size
	SWAP		d4				* use the remainder
	BSR		get_rows			* get the number of screen rows in d1.w
	ADD.w		d1,d4				* add the number of rows
	BRA.s		erase_tail			* go do erase tail

erase_loop
	MOVEQ		#0,d1				* clear longword
	MOVE.b	d5,d1				* copy next y%
	BMI		erase_character		* if next y% < 0 go erase character

	SUB.w		d4,d1				* subtract random y%
	BMI		erase_character		* if erase y% > next go erase character

	CMP.w		#tail,d1			* compare distance with tail length
	BCS		erase_tail			* if erase y% < tail from next skip erase

erase_character
	MOVE.b	d7,d1				* copy column number
	ASL.w		#8,d1				* shift to high byte
	OR.b		d4,d1				* OR in random row
	MOVEQ		#11,d0			* position the cursor at row, column
	TRAP		#15

	MOVEQ		#' ',d1			* set [SPACE] character
	MOVEQ		#6,d0				* display single character in d1.b
	TRAP		#15

erase_tail
	SUB.w		#erase,d4			* index to next erase row
	BPL		erase_loop			* go erase another character

	RTS


*************************************************************************************
*
* get the number of screen rows in d1.w

get_rows
	MOVEQ		#0,d1				* clear longword
	MOVE.w	screen_height(a3),d1	* get screen height
	DIVU.w	#15,d1			* divide by character height
	RTS


*************************************************************************************
*
* print the character and colour it, enter with y index in d2.w

colour_character
	BSR		random_n			* generate a random number
	MOVEQ		#$7F,d1			* set the character index mask
	AND.b		(a3),d1			* AND with a random byte
	MOVE.b	(a1,d1.w),d1		* get a character from the table
	MOVEQ		#6,d0				* display single character in d1.b
	TRAP		#15

	ASL.w		#2,d4				* * 4 bytes per colour longword
	MOVE.l	(a2,d4.w),d1		* get colour(colour%)
	MOVEQ		#80,d0			* set the pen colour
	TRAP		#15

	MOVEQ		#81,d0			* set the fill colour
	TRAP		#15

* character is printed in white, now colour it by drawing a filled rectangle over it in
* AND mode

	MOVE.w	d7,d1				* set x origin
	ASL.w		#3,d1				* x * 8 pixels
	MULU.w	#$0F,d2			* y * 15 pixels
	MOVEQ		#$07,d3			* set 8 pixels wide
	MOVEQ		#$0E,d4			* set 15 pixels high
	ADD.w		d1,d3				* add x origin
	ADD.w		d2,d4				* add y origin
	MOVEQ		#87,d0			* draw a filled rectangle
	TRAP		#15

	MOVEQ		#0,d1				* set $00000000, black fill
	MOVEQ		#81,d0			* set the fill colour
	TRAP		#15				*

	RTS


*************************************************************************************
*
* set the text cursor to (column%, d2)

set_cursor
	MOVE.b	d7,d1				* copy column number
	ASL.w		#8,d1				* shift to high byte
	OR.b		d2,d1				* OR in row
	MOVEQ		#11,d0			* position the cursor at row, column
	TRAP		#15

	RTS


*************************************************************************************
*
* initialise everything

init
	LEA		row(pc),a6			* row.w(column%)
	LEA		deltav(pc),a5		* deltav.w(column%)
	LEA		colour(pc),a4		* colour.b(column%)
	LEA		PRNlword(pc),a3		* random seed
	LEA		palette(pc),a2		* colour palette table
	LEA		text(pc),a1			* text$(string.b(column%))

	MOVE.l	#$31415927,(a3)		* set PRNG seed. this can be any value but 0
	MOVE.l	#$028001E0,d1		* set default window size to 640 x 480

restart_screen
	MOVE.w	#$FE00,d0			* clear row #
	MOVE.w	#$FE,d7			* set end column
init_loop1
	MOVE.w	d0,(a6,d7.w)		* clear row.w(d7)
	SUBQ.w	#2,d7				* decrement column ..
	BPL		init_loop1			* .. and loop if more to do

	MOVE.l	d1,screen_width(a3)	* save the window size
	MOVEQ 	#33,d0			* get/set window size
	TRAP		#15

	RTS


*************************************************************************************
*
* This is the code that generates the pseudo random sequence. A seed word located in
* PRNlword(a3) is loaded into a register before being operated on to generate the
* next number in the sequence. This number is then saved as the seed for the next
* time it's called.
*
* This code is adapted from the 32 bit version of RND(n) used in EhBASIC68. Taking
* the 19th next number is slower but helps to hide the shift and add nature of this
* generator as can be seen from analysing the output.

random_n
	MOVEM.l	d0-d2,-(sp)			* save d0, d1 and d2
	MOVE.l	(a3),d0			* get current seed longword
	MOVEQ		#$AF-$100,d1		* set EOR value
	MOVEQ		#18,d2			* do this 19 times
Ninc0
	ADD.l		d0,d0				* shift left 1 bit
	BCC.s		Ninc1				* if bit not set skip feedback

	EOR.b		d1,d0				* do Galois LFSR feedback
Ninc1
	DBF		d2,Ninc0			* loop

	MOVE.l	d0,(a3)			* save back to seed longword
	MOVEM.l	(sp)+,d0-d2			* restore d0, d1 and d2

	RTS


*************************************************************************************
*
* the palette. the last colour is the string head character colour. the rest are
* shades of green for the string tails

palette
	dc.l	$00003000
	dc.l	$00003800
	dc.l	$00004000
	dc.l	$00004800
	dc.l	$00005000
	dc.l	$00005800
	dc.l	$00006000
	dc.l	$00006800
	dc.l	$00007000
	dc.l	$00007800
	dc.l	$00008000
	dc.l	$00008800
	dc.l	$00009000
	dc.l	$00009800
	dc.l	$0000A000
	dc.l	$0080FF80


*************************************************************************************
*
* character table, random characters are selected from this

text
	dc.b	'Lee_Davison-28/O1\20O7=V8$qk5"}`'
	dc.b	'ujz]Z(!m/?%HyZ%4@(BCGHIJMPQUVXY#'
	dc.b	'Zbcdfghjlmpquwxz3459!$%^&*+=[]|'
	dc.b	'WriTteN.FoR,EASky68K~{}:@;"<>?()'
row
	ds.w	$80					* column vertical position, real 8.8
deltav
	ds.w	$80					* column velocity, real 8.8
string
	ds.b	$80					* column index, integer byte
colour
	ds.b	$80					* column colour, integer byte
PRNlword
	ds.l	1					* PRNG seed long word

screen_width
	ds.w	1					* screen width
screen_height
	ds.w	1					* screen height

	END	start


*************************************************************************************

